/*-
 * Copyright (C) 2012 Oleg Moskalenko <oleg.moskalenko@citrix.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "radixsort.h"
#include "coll.h"

static bool reverse_sort = false;

static const size_t slsz = 256 * sizeof(struct sort_level*);

struct sort_level
{

	size_t			level;

	struct sort_list_item	**tosort;
	size_t			tosort_num;
	size_t			tosort_sz;

	struct sort_list_item	**leaves;
	size_t			leaves_num;
	size_t			leaves_sz;

	struct sort_level	**sublevels;
	size_t			sln;
	size_t			real_sln;
};

static void
add_to_sublevel(struct sort_level *sl, struct sort_list_item *item, int indx)
{
	struct sort_level *ssl;

	ssl = sl->sublevels[indx];

	if (ssl == NULL) {
		ssl = sort_malloc(sizeof(struct sort_level));
		if (ssl == NULL)
			err(2, NULL);
		memset(ssl, 0, sizeof(struct sort_level));

		ssl->level = sl->level + 1;
		sl->sublevels[indx] = ssl;

		++(sl->real_sln);
	}

	if (++(ssl->tosort_num) > ssl->tosort_sz) {
		ssl->tosort_sz = ssl->tosort_num + 128;
		ssl->tosort = sort_realloc(ssl->tosort, sizeof(struct sort_list_item*) * (ssl->tosort_sz));
	}

	ssl->tosort[ssl->tosort_num - 1] = item;
}

static inline void
add_leaf(struct sort_level *sl, struct sort_list_item *item)
{
	if (++(sl->leaves_num) > sl->leaves_sz) {
		sl->leaves_sz = sl->leaves_num + 128;
		sl->leaves = sort_realloc(sl->leaves, (sizeof(struct sort_list_item*) * (sl->leaves_sz)));
	}
	sl->leaves[sl->leaves_num - 1] = item;
}

static inline int
get_wc_index(struct sort_list_item *sli, size_t level)
{
	const struct bwstring *bws;

	bws = sli->ka.key[0].k;

	if ((BWSLEN(bws) > level))
		return (unsigned char) BWS_GET(bws,level);
	return -1;
}

static void
place_item(struct sort_level *sl, size_t item)
{
	struct sort_list_item *sli;
	int c;

	sli = sl->tosort[item];
	c = get_wc_index(sli, sl->level);

	if (c == -1)
		add_leaf(sl, sli);
	else
		add_to_sublevel(sl, sli, c);
}

static void
free_sort_level(struct sort_level *sl)
{
	if (sl) {
		if (sl->leaves)
			sort_free(sl->leaves);
		if (sl->level > 0 && sl->tosort)
			sort_free(sl->tosort);
		if (sl->sublevels) {

			size_t i;

			for (i = 0; i < sl->sln; ++i)
				free_sort_level(sl->sublevels[i]);
			sort_free(sl->sublevels);
		}
		sort_free(sl);
	}
}

static void
run_sort_level_next(struct sort_level *sl, struct sort_list_item **sorted, size_t *sorted_n)
{
	size_t i;
	size_t tosort_num;
	size_t sln;

	if (sl->sublevels) {
		sort_free(sl->sublevels);
		sl->sublevels = NULL;
	}

	switch (sl->tosort_num){
	case 0:
		if (sl->tosort) {
			sort_free(sl->tosort);
			sl->tosort = NULL;
		}
		return;
	case (1):
		sorted[(*sorted_n)++] = sl->tosort[0];
		sort_free(sl->tosort);
		sl->tosort = NULL;
		return;
	case (2):
		if (list_coll_offset(&(sl->tosort[0]), &(sl->tosort[1]), sl->level) > 0) {
			sorted[(*sorted_n)++] = sl->tosort[1];
			sorted[(*sorted_n)++] = sl->tosort[0];
		} else {
			sorted[(*sorted_n)++] = sl->tosort[0];
			sorted[(*sorted_n)++] = sl->tosort[1];
		}
		sort_free(sl->tosort);
		sl->tosort = NULL;
		return;
	default:
		if ((sl->tosort_num < 65) || (sl->level > 15)) {

			listcoll_t func;

			func = get_list_call_func(sl->level);

			sl->leaves = sl->tosort;
			sl->leaves_num = sl->tosort_num;
			sl->leaves_sz = sl->leaves_num;
			sl->leaves = sort_realloc(sl->leaves, (sizeof(struct sort_list_item*) * (sl->leaves_sz)));
			sl->tosort = NULL;
			sl->tosort_num = 0;
			sl->tosort_sz = 0;
			sl->sln = 0;
			sl->real_sln = 0;
			if (sort_opts_vals.sflag) {
				if (mergesort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) func) == -1)
					err(2, "Radix sort error 3");
			} else
				qsort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) func);

			memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num * sizeof(struct sort_list_item*));
			(*sorted_n) += sl->leaves_num;

			sort_free(sl->leaves);
			sl->leaves = NULL;
			sl->leaves_num = 0;
			sl->leaves_sz = 0;

			return;
		} else {
			sl->tosort_sz = sl->tosort_num;
			sl->tosort = sort_realloc(sl->tosort, sizeof(struct sort_list_item*) * (sl->tosort_sz));
		}
	}

	sl->sln = 256;
	sl->sublevels = sort_malloc(slsz);
	memset(sl->sublevels, 0, slsz);

	sl->real_sln = 0;

	tosort_num = sl->tosort_num;
	for (i = 0; i < tosort_num; ++i)
		place_item(sl, i);

	sort_free(sl->tosort);
	sl->tosort = NULL;
	sl->tosort_num = 0;
	sl->tosort_sz = 0;

	if (sl->leaves_num > 1) {
		if (keys_num > 1) {
			if (sort_opts_vals.sflag) {
				mergesort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) list_coll);
			} else {
				qsort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) list_coll);
			}
		} else if (!sort_opts_vals.sflag && sort_opts_vals.complex_sort) {
			qsort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(const void *,
							const void *)) list_coll_by_str_only);
		}
	}

	sl->leaves_sz = sl->leaves_num;
	sl->leaves = sort_realloc(sl->leaves, (sizeof(struct sort_list_item*) * (sl->leaves_sz)));

	if (!reverse_sort) {

		memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num * sizeof(struct sort_list_item*));
		(*sorted_n) += sl->leaves_num;

		sort_free(sl->leaves);
		sl->leaves = NULL;
		sl->leaves_num = 0;
		sl->leaves_sz = 0;

		sln = sl->sln;
		for (i = 0; i < sln; ++i) {

			struct sort_level *slc;

			slc = sl->sublevels[i];

			if (slc) {
				run_sort_level_next(slc, sorted, sorted_n);
				free_sort_level(slc);
				sl->sublevels[i] = NULL;
			}
		}

	} else {

		sln = sl->sln;
		for (i = 0; i < sln; ++i) {

			size_t n;
			struct sort_level *slc;

			n = sln - i - 1;
			slc = sl->sublevels[n];

			if (slc) {
				run_sort_level_next(slc, sorted, sorted_n);
				free_sort_level(slc);
				sl->sublevels[n] = NULL;
			}
		}

		memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num * sizeof(struct sort_list_item*));
		(*sorted_n) += sl->leaves_num;

		sort_free(sl->leaves);
		sl->leaves = NULL;
		sl->leaves_num = 0;
		sl->leaves_sz = 0;
	}

	if (sl->sublevels) {
		sort_free(sl->sublevels);
		sl->sublevels = NULL;
		sl->sln = 0;
	}

}

static void
run_top_sort_level(struct sort_level *sl)
{
	size_t i;
	size_t sorted_n;

	if (sort_opts_vals.kflag) {
		reverse_sort = keys[0].sm.rflag;
	} else {
		reverse_sort = default_sort_mods->rflag;
	}

	i = 0;
	sorted_n = 0;

	sl->sln = 256;
	sl->sublevels = sort_malloc(slsz);
	memset(sl->sublevels, 0, slsz);

	for (i = 0; i < sl->tosort_num; ++i)
		place_item(sl, i);

	if (sl->leaves_num > 1) {
		if (keys_num > 1) {
			if (sort_opts_vals.sflag) {
				mergesort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) list_coll);
			} else {
				qsort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(
								const void *, const void *)) list_coll);
			}
		} else if (!sort_opts_vals.sflag && sort_opts_vals.complex_sort) {
			qsort(sl->leaves, sl->leaves_num, sizeof(struct sort_list_item *), (int(*)(const void *,
							const void *)) list_coll_by_str_only);
		}
	}

	if (!reverse_sort) {
		memcpy(sl->tosort + sorted_n, sl->leaves, sl->leaves_num * sizeof(struct sort_list_item*));
		sorted_n += sl->leaves_num;
		for (i = 0; i < sl->sln; ++i) {

			struct sort_level *slc;

			slc = sl->sublevels[i];

			if (slc) {
				run_sort_level_next(slc, sl->tosort, &sorted_n);
				free_sort_level(slc);
				sl->sublevels[i] = NULL;
			}
		}
	} else {
		for (i = 0; i < sl->sln; ++i) {

			size_t n;
			struct sort_level *slc;

			n = sl->sln - i - 1;
			slc = sl->sublevels[n];

			if (slc) {
				run_sort_level_next(slc, sl->tosort, &sorted_n);
				free_sort_level(slc);
				sl->sublevels[n] = NULL;
			}
		}
		memcpy(sl->tosort + sorted_n, sl->leaves, sl->leaves_num * sizeof(struct sort_list_item*));
		sorted_n += sl->leaves_num;
	}
}

static void
run_sort(struct sort_list_item **base, size_t nmemb)
{
	struct sort_level *sl;

	sl = sort_malloc(sizeof(struct sort_level));
	if (sl == NULL)
		err(2, NULL);
	memset(sl, 0, sizeof(struct sort_level));

	sl->tosort = base;
	sl->tosort_num = nmemb;
	sl->tosort_sz = nmemb;

	run_top_sort_level(sl);

	free_sort_level(sl);
}

void
rxsort(struct sort_list_item **base, size_t nmemb)
{
	run_sort(base, nmemb);
}
